#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from grib_check.Assert import Fail, Pass
from grib_check.Report import Report


class TestAssert:

    def test_level_1(self):
        report = Report("Report 1")
        report.add(Pass("Assertion 1"))
        assert str(report) == "PASS: Report 1\n  PASS: Assertion 1\n"

        report.add(Fail("Assertion 2"))
        assert (
            str(report) == "FAIL: Report 1\n  PASS: Assertion 1\n  FAIL: Assertion 2\n"
        )

    def test_nested_report(self):
        report = Report("Report 1")
        report.add(Pass("Assertion 1"))

        nested = Report("Report 2")
        nested.add(Pass("Assertion 2"))

        report.add(nested)

        assert (
            str(report)
            == "PASS: Report 1\n  PASS: Assertion 1\n  PASS: Report 2\n    PASS: Assertion 2\n"
        )

    def test_assertions(self):
        report1 = Report("Test report")
        report1.add(Fail("Some error message"))
        assert str(report1) == "FAIL: Test report\n  FAIL: Some error message\n"

        f = Fail("Some error message")
        p = Pass("Some success message")
        report2 = Report("Test report 2")
        report2.add(f & p)

        assert (
            str(report2)
            == "FAIL: Test report 2\n  FAIL: Some error message and Some success message\n"
        )

    def test_add_string(self):
        report = Report("Test report")
        report.add(Fail("Some error message"))
        report.add("A message")

        assert (report.status() is False)

    def test_add_failed_report(self):
        report1 = Report("Report 1")
        report1.add(Fail("Bad message"))
        report2 = Report("Report 2")
        report2.add(Pass("Good message"))

        report = Report("Top report")
        report.add(report1)
        report.add(report2)

        assert (report.status() is False)



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import json
import sys

# import pandas as pd
import _jsonnet

from .Message import Message
from .Report import Report


class LookupTable:
    def __init__(self, filename: str):
        raise NotImplementedError

    def get_element(self, message: Message):
        raise NotImplementedError


class SimpleLookupTable(LookupTable):
    """
    A simple lookup table that uses a JSON file to store the data.
    """

    def __init__(self, filename: str, ignore_keys=None):
        assert filename is not None
        try:
            jresult = _jsonnet.evaluate_file(filename)
            self.data = json.loads(jresult)
        except ValueError as e:
            print(f"ERROR: Couldn't read JSON file {filename}: {e}", file=sys.stderr)
            sys.exit(1)

        self.ignore_keys = ignore_keys

    def get_element(self, message: Message):
        report = Report("Matched parameter")
        params = list()
        for row in self.data:
            count = 0
            count_ignore = 0
            for pair in row["pairs"]:
                if self.ignore_keys is not None and pair["key"] in self.ignore_keys:
                    count_ignore += 1
                    continue
                if message.get(pair["key"], type(pair["value"])) == pair["value"]:
                    count += 1
            if count == len(row["pairs"]) - count_ignore:
                params.append((count, row))
        if len(params) > 0:
            params.sort(key=lambda x: x[0], reverse=True)
            if "name" in params[0][1]:
                report.rename(f"Matched parameter: {params[0][1]['name']}")
            for pair in params[0][1]["pairs"]:
                report.add(pair["key"] + ": " + str(pair["value"]))
            return params[0][1], report

        return (None, report)


class IndexedLookupTable(LookupTable):
    """
    A lookup table that uses a dictionary to store the data.
    The dictionary is indexed by the keys in the data.

    TODO: Implement and replace SimpleLookupTable with this class.
    """

    def __init__(self, filename: str):
        raise NotImplementedError

    def get_element(self, message: Message):
        raise NotImplementedError



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from __future__ import annotations

import logging
from enum import Enum

from .ValueFormat import formatter


class OpType(Enum):
    DASH = 0
    DOT = 1
    NONE = 2


def makeKV(value):
    if type(value) is KeyValue:
        return value
    else:
        return KeyValue(None, value)


class KeyValue:
    def __init__(self, key, value, level=0, last_op_type=OpType.NONE, key_suffix=None):
        self.logger = logging.getLogger(__class__.__name__)
        self.__key = key
        self.__value = value
        self.__level = level
        self.__last_op_type = last_op_type
        self.__key_suffix = key_suffix  # e.g. "[0]" in key[0](value)

    def key(self):
        return self.__key

    def value(self):
        return self.__value

    def type(self):
        return type(self.__value)

    def __str__(self):
        if self.__key is None:
            return formatter.format(self.__value)
        else:
            key_str = f'{self.__key}{self.__key_suffix if self.__key_suffix is not None else ""}'
            if self.__level == 0:
                return f"{key_str}({formatter.format(self.__value)})"
            else:
                return f"{key_str}"

    def __repr__(self):
        if self.__level == 0:
            return f"{self.__key}({self.__value})"
        else:
            return f"{self.__key}"

    def __eq__(self, other: KeyValue) -> bool:
        otherKV = makeKV(other)
        return bool(self.__value == otherKV.__value)

    def __ne__(self, other: KeyValue) -> bool:
        otherKV = makeKV(other)
        return bool(self.__value != otherKV.__value)

    def __lt__(self, other: KeyValue) -> bool:
        otherKV = makeKV(other)
        return self.__value < otherKV.__value

    def __le__(self, other: KeyValue) -> bool:
        otherKV = makeKV(other)
        return self.__value <= otherKV.__value

    def __gt__(self, other: KeyValue) -> bool:
        otherKV = makeKV(other)
        return self.__value > otherKV.__value

    def __ge__(self, other: KeyValue) -> bool:
        otherKV = makeKV(other)
        return self.__value >= otherKV.__value

    def __dash_op__(self, lhs: KeyValue, rhs: KeyValue, op, op_sign, op_type) -> KeyValue:
        assert type(lhs) is KeyValue and type(rhs) is KeyValue, f"lhs and rhs must be KeyValue, got {type(lhs)} and {type(rhs)}"
        lhs_last_op_type = lhs.__last_op_type
        rhs_last_op_type = rhs.__last_op_type

        lhss = (
            f"({lhs})"
            if lhs_last_op_type == OpType.DASH and op_type == OpType.DOT
            else f"{lhs}"
        )
        rhss = (
            f"({rhs})"
            if rhs_last_op_type == OpType.DASH and op_type == OpType.DOT
            else f"{rhs}"
        )

        k = f"{lhss} {op_sign} {rhss}"

        if lhs is None or rhs is None:
            return KeyValue(k, None, self.__level + 1)

        if type(lhs) is KeyValue and lhs.__value is None:
            return KeyValue(k, None, self.__level + 1)

        if type(rhs) is KeyValue and rhs.__value is None:
            return KeyValue(k, None, self.__level + 1)

        v1 = lhs.__value
        v2 = rhs.__value

        return KeyValue(k, op(v1, v2), self.__level + 1, op_type)

    def __add__(self, other: KeyValue) -> KeyValue:
        return self.__dash_op__(self, makeKV(other), lambda x, y: x + y, "+", OpType.DASH)

    def __radd__(self, other: KeyValue) -> KeyValue:
        return self.__dash_op__(makeKV(other), self, lambda x, y: x + y, "+", OpType.DASH)

    def __sub__(self, other: KeyValue) -> KeyValue:
        return self.__dash_op__(self, makeKV(other), lambda x, y: x - y, "-", OpType.DASH)

    def __rsub__(self, other: KeyValue) -> KeyValue:
        return self.__dash_op__(makeKV(other), self, lambda x, y: x - y, "-", OpType.DASH)

    def __mul__(self, other: KeyValue) -> KeyValue:
        return self.__dash_op__(self, makeKV(other), lambda x, y: x * y, "*", OpType.DOT)

    def __rmul__(self, other: KeyValue) -> KeyValue:
        return self.__dash_op__(makeKV(other), self, lambda x, y: x * y, "*", OpType.DOT)

    def __truediv__(self, other: KeyValue) -> KeyValue:
        return self.__dash_op__(self, makeKV(other), lambda x, y: x / y, "/", OpType.DOT)

    def __rtruediv__(self, other: KeyValue) -> KeyValue:
        return self.__dash_op__(makeKV(other), self, lambda x, y: x / y, "/", OpType.DOT)

    def __mod__(self, other) -> KeyValue:
        k = f"{self} % {other}" if self.__level == 0 else f"({self}) % {other}"
        if not (
            isinstance(self.__value, (int, float)) and isinstance(other, (int, float))
        ):
            return KeyValue(k, None, self.__level + 1)
        v = None if self.__value is None else self.__value % other
        return KeyValue(k, v, self.__level + 1)

    def __neg__(self):
        k = f"-({self})" if self.__level == 0 else f"-({self})"
        v = None if self.__value is None else -self.__value
        return KeyValue(k, v, self.__level + 1)

    def __rcontains__(self, other):
        otherKV = makeKV(other)
        k = f"{self} in {otherKV.__value}"
        v = self.__value in otherKV.__value
        return KeyValue(k, v, self.__level + 1)

    def __int__(self):
        return int(self.__value)

    def abs(self):
        k = f"|{self}|"
        v = None if self.__value is None else abs(self.__value)
        return KeyValue(k, v, self.__level + 1)

    def to_int(self):
        k = f"int({self})"
        v = None if self.__value is None else int(self.__value)
        return KeyValue(k, v, self.__level + 1)



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from grib_check.Assert import Eq
from grib_check.Grib import Grib
from grib_check.KeyValue import KeyValue


class TestAssert:
    def test_and(self):
        kv1 = KeyValue("stream", "eefo")
        kv2 = KeyValue("stream", "nai")

        eq1 = Eq(kv1, kv1)
        eq2 = Eq(kv1, kv2)

        assert eq1
        assert not eq2

        a = eq1 & eq2
        assert not a
        assert f"{eq1} and {eq2}" == a.as_string()

        a = eq1 & eq1
        assert a
        assert f"{eq1} and {eq1}" == a.as_string()

    def test_or(self):
        kv1 = KeyValue("stream", "eefo")
        kv2 = KeyValue("stream", "nai")

        eq1 = Eq(kv1, kv1)
        eq2 = Eq(kv1, kv2)

        assert eq1
        assert not eq2

        a = eq1 | eq2
        assert a
        assert f"{eq1} or {eq2}" == a.as_string()

        a = eq1 | eq1
        assert a
        assert f"{eq1} or {eq1}" == a.as_string()

    def test_eq(self):
        from grib_check.Assert import Eq

        kv1 = KeyValue("stream", "eefo")
        kv2 = KeyValue("stream", "nai")

        eq = Eq(kv1, kv1)
        assert eq
        assert f"{kv1} == {kv1}" == eq.as_string()

        eq = Eq(kv1, kv2)
        assert not eq
        assert f"{kv1} == {kv2}" == eq.as_string()

        eq = Eq(kv1, "eefo")
        assert eq
        assert f"{kv1} == eefo" == eq.as_string()

        eq = Eq("eefo", kv1)
        assert eq
        assert f"eefo == {kv1}" == eq.as_string()

        eq = Eq(kv2, "nai")
        assert eq
        assert f"{kv2} == nai" == eq.as_string()

        eq = Eq("nai", kv2)
        assert eq
        assert f"nai == {kv2}" == eq.as_string()

        eq = Eq("eefo", "eefo")
        assert eq
        assert "eefo == eefo" == eq.as_string()

        eq = Eq("eefo", "nai")
        assert not eq
        assert "eefo == nai" == eq.as_string()

    def test_isin(self):
        from grib_check.Assert import IsIn

        kv = KeyValue("stream", "eefo")

        isin = IsIn(kv, ["eefo", "nai"])
        assert isin
        assert "stream(eefo) in ['eefo', 'nai']" == isin.as_string()

        isin = IsIn(kv, ["nai", "dgov"])
        assert not isin
        assert "stream(eefo) in ['nai', 'dgov']" == isin.as_string()

    def test_multiple_of(self):
        from grib_check.Assert import IsMultipleOf

        multiple_of = IsMultipleOf(KeyValue("test", 12), 6)
        assert multiple_of
        assert "test(12) % 6 == 0" == multiple_of.as_string()

        multiple_of = IsMultipleOf(KeyValue("test", 12), 7)
        assert not multiple_of
        assert "test(12) % 7 == 0" == multiple_of.as_string()

    def test_exists(self):
        from grib_check.Assert import Exists

        grib = Grib("tests/dgov-data/od_eefo_fcmean_sfc_2024_0001_reduced_gg.grib2")
        message = grib.__next__()

        exists = Exists(message, "stream")
        assert exists
        assert (
            "stream exists(True) and has a non-missing value(True)"
            == exists.as_string()
        )

        exists = Exists(message, "non_existent_key")
        assert not exists
        assert (
            "non_existent_key exists(False) and has a non-missing value(None)"
            == exists.as_string()
        )

        exists = Exists(message, "hoursAfterDataCutoff")
        assert not exists
        assert (
            "hoursAfterDataCutoff exists(True) and has a non-missing value(False)"
            == exists.as_string()
        )

    def test_missing(self):
        from grib_check.Assert import Missing

        grib = Grib("tests/dgov-data/od_eefo_fcmean_sfc_2024_0001_reduced_gg.grib2")
        message = grib.__next__()

        missing = Missing(message, "stream")
        assert not missing
        assert (
            "stream exists(True) and has a missing value(False)" == missing.as_string()
        )

        missing = Missing(message, "non_existent_key")
        assert not missing
        assert (
            "non_existent_key exists(False) and has a missing value(None)"
            == missing.as_string()
        )

        missing = Missing(message, "hoursAfterDataCutoff")
        assert (
            "hoursAfterDataCutoff exists(True) and has a missing value(True)"
            == missing.as_string()
        )
        assert missing

    def test_eq_double(self):
        from grib_check.Assert import EqDbl

        kv = KeyValue("test", 6.0001)
        eq = EqDbl(kv, 6.0, 0.01)
        assert eq
        assert "test(6.0001) == 6.0 within 0.01" == eq.as_string()

        eq = EqDbl(kv, 6.0, 0.00001)
        assert not eq
        assert "test(6.0001) == 6.0 within 1e-05" == eq.as_string()

        eq = EqDbl(kv, 7.0, 0.01)
        assert not eq
        assert "test(6.0001) == 7.0 within 0.01" == eq.as_string()

        # TODO(maee): Add functionality to compare KeyValue with KeyValue

        # eq = EqDbl(6.0001, 6.0, 0.01)
        # assert eq
        # assert "6.0001 == 6.0 within 0.01" == eq.as_string()

        # eq = EqDbl(6, kv, 0.00001)
        # assert not eq
        # assert "6 == test(6.0001) within 1e-05" == eq.as_string()

        # eq = EqDbl(6, 6, 0.01)
        # assert eq
        # assert "6 == 6 within 0.01" == eq.as_string()

    def test_ne(self):
        from grib_check.Assert import Ne

        kv1 = KeyValue("test", 6)
        kv2 = KeyValue("test", 7)
        ne = Ne(kv1, kv2)
        assert ne
        assert "test(6) != test(7)" == ne.as_string()

        ne = Ne(kv1, 6)
        assert not ne
        assert "test(6) != 6" == ne.as_string()

        ne = Ne(kv1, 7)
        assert ne
        assert "test(6) != 7" == ne.as_string()

        ne = Ne(6, 7)
        assert ne
        assert "6 != 7" == ne.as_string()

        ne = Ne(6, 6)
        assert not ne
        assert "6 != 6" == ne.as_string()

        ne = Ne(6, kv1)
        assert not ne
        assert "6 != test(6)" == ne.as_string()

        ne = Ne(7, kv1)
        assert ne
        assert "7 != test(6)" == ne.as_string()

    def test_ge(self):
        from grib_check.Assert import Ge

        kv1 = KeyValue("test", 6)
        kv2 = KeyValue("test", 7)
        ge = Ge(kv1, kv2)
        assert not ge
        assert "test(6) >= test(7)" == ge.as_string()

        ge = Ge(kv1, 6)
        assert ge
        assert "test(6) >= 6" == ge.as_string()

        ge = Ge(kv1, 5)
        assert ge
        assert "test(6) >= 5" == ge.as_string()

        ge = Ge(6, 7)
        assert not ge
        assert "6 >= 7" == ge.as_string()

        ge = Ge(6, 6)
        assert ge
        assert "6 >= 6" == ge.as_string()

        ge = Ge(7, kv1)
        assert ge
        assert "7 >= test(6)" == ge.as_string()

        ge = Ge(5, kv1)
        assert not ge
        assert "5 >= test(6)" == ge.as_string()

    def test_le(self):
        from grib_check.Assert import Le

        kv1 = KeyValue("test", 6)
        kv2 = KeyValue("test", 7)
        le = Le(kv1, kv2)
        assert le
        assert "test(6) <= test(7)" == le.as_string()

        le = Le(kv1, 6)
        assert le
        assert "test(6) <= 6" == le.as_string()

        le = Le(kv1, 5)
        assert not le
        assert "test(6) <= 5" == le.as_string()

        le = Le(6, 7)
        assert le
        assert "6 <= 7" == le.as_string()

        le = Le(6, 6)
        assert le
        assert "6 <= 6" == le.as_string()

        le = Le(7, kv1)
        assert not le
        assert "7 <= test(6)" == le.as_string()

        le = Le(5, kv1)
        assert le
        assert "5 <= test(6)" == le.as_string()

    def test_gt(self):
        from grib_check.Assert import Gt

        kv1 = KeyValue("test", 6)
        kv2 = KeyValue("test", 7)
        gt = Gt(kv1, kv2)
        assert not gt
        assert "test(6) > test(7)" == gt.as_string()

        gt = Gt(kv1, 6)
        assert not gt
        assert "test(6) > 6" == gt.as_string()

        gt = Gt(kv1, 5)
        assert gt
        assert "test(6) > 5" == gt.as_string()

        gt = Gt(6, 7)
        assert not gt
        assert "6 > 7" == gt.as_string()

        gt = Gt(6, 6)
        assert not gt
        assert "6 > 6" == gt.as_string()

        gt = Gt(7, kv1)
        assert gt
        assert "7 > test(6)" == gt.as_string()

        gt = Gt(5, kv1)
        assert not gt
        assert "5 > test(6)" == gt.as_string()

    def test_lt(self):
        from grib_check.Assert import Lt

        kv1 = KeyValue("test", 6)
        kv2 = KeyValue("test", 7)
        lt = Lt(kv1, kv2)
        assert lt
        assert "test(6) < test(7)" == lt.as_string()

        lt = Lt(kv1, 6)
        assert not lt
        assert "test(6) < 6" == lt.as_string()

        lt = Lt(kv1, 5)
        assert not lt
        assert "test(6) < 5" == lt.as_string()

        lt = Lt(6, 7)
        assert lt
        assert "6 < 7" == lt.as_string()

        lt = Lt(6, 6)
        assert not lt
        assert "6 < 6" == lt.as_string()

        lt = Lt(7, kv1)
        assert not lt
        assert "7 < test(6)" == lt.as_string()

        lt = Lt(5, kv1)
        assert lt
        assert "5 < test(6)" == lt.as_string()

    def test_fail(self):
        from grib_check.Assert import Fail

        fail = Fail("This is a failure")
        assert not fail
        assert "This is a failure" == fail.as_string()

        fail = Fail("Another failure")
        assert not fail
        assert "Another failure" == fail.as_string()

    def test_pass(self):
        from grib_check.Assert import Pass

        pas = Pass("This is a pass")
        assert pas
        assert "This is a pass" == pas.as_string()

        pas = Pass("Another pass")
        assert pas
        assert "Another pass" == pas.as_string()



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from grib_check.Assert import IsIn, IsMultipleOf, Le
from grib_check.Report import Report

from .GeneralChecks import GeneralChecks


class DestinE(GeneralChecks):
    def __init__(self, lookup_table, valueflg=False):
        super().__init__(lookup_table, valueflg=valueflg)
        self.register_checks({"destine_limits": self._destine_limits})

    # Reuse / override checks
    def _point_in_time(self, message, p) -> Report:
        report = Report("Point In Time (DestinE)")
        report.add(IsMultipleOf(message["step"], 3))
        return super()._point_in_time(message, p).add(report)

    # Create new checks
    def _destine_limits(self, message, p) -> Report:
        report = Report("DestinE Limits")
        report.add(Le(message["step"], 30))
        report.add(IsIn(message["forecastTime"], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]))
        report.add(IsIn(message["indicatorOfUnitOfTimeRange"], [0, 1]))
        report.add(IsIn(message["timeIncrementBetweenSuccessiveFields"], [0, 1]))
        return report



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from grib_check.Grib import Grib
from grib_check.LookupTable import SimpleLookupTable


class TestIndexedLookupTable:
    def test_get_element(self):
        table = SimpleLookupTable("tests/test_parameters.json")
        grib = Grib("tests/wmo/od_eefo_fcmean_sfc_2024_0001_reduced_gg.grib2")
        message = next(grib)
        element, lookup_table_report = table.get_element(message)
        expected = {
            "name": "param_1",
            "pairs": [
                {"key": "stream", "value": "eefo"},
                {"key": "dataType", "value": "fcmean"},
            ],
            "expected": [
                {"key": "productDefinitionTemplateNumber", "value": 11},
                {"key": "paramId", "value": "228004"},
                {"key": "shortName", "value": "mean2t"},
                {"key": "name", "value": "Mean 2 metre temperature"},
            ],
            "checks": ["basic_checks"],
        }
        assert expected == element



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from .Message import Message
from .Report import Report


class Test:
    """
    This class determines the order in which the checks are executed.
    As a rule, it receives a "parameter" with a list of the checks to be executed and a "check_map"
    indicating where the code for the respective tests is located.
    The test is then applied to the "message".
    """

    def __init__(self, message: Message, parameter: dict, check_map: dict):
        raise NotImplementedError

    def run(self) -> Report:
        raise NotImplementedError



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from grib_check.Assert import AssertTrue, Eq, Fail, IsIn, IsMultipleOf, Le
from grib_check.Report import Report

from .GeneralChecks import GeneralChecks


class Uerra(GeneralChecks):
    def __init__(self, lookup_table, check_limits=False, check_validity=True):
        super().__init__(lookup_table, check_limits=check_limits, check_validity=check_validity)

    def _basic_checks_2(self, message, p) -> Report:
        # this class must not inhereted anything
        report = Report("UERRA Basic Checks 2")

        report.add(IsIn(message["productionStatusOfProcessedData"], [8, 9]))
        report.add(Le(message["endStep"], 30))
        report.add(
            IsIn(message["typeOfProcessedData"], [0, 1])
        )  # 0 = analysis, 1 = forecast
        if message["typeOfProcessedData"] == 0:
            report.add(Eq(message["step"], 0))
        else:
            report.add(IsIn(message["step"], [1, 2, 4, 5]) | IsMultipleOf(message["step"], 3))
        report.add(self._check_date(message, p))

        return report

    def _basic_checks(self, message, p) -> Report:
        report = Report("UERRA Basic Checks")
        if message.get("class", str) != 'rr' and message.get("class", str) != 'ci':
            report.add(Eq(message["versionNumberOfGribLocalTables"], 0))

        report.add(Le(message["hour"], 24))
        stream = message.get("stream", str)

        if stream != "moda":
            report.add(IsIn(message["step"], [1, 2, 4, 5]) | IsMultipleOf(message["step"], 3))

        return super()._basic_checks(message, p).add(report)

    def _from_start(self, message, p) -> Report:
        report = Report("Uerra From Start")
        min_value, max_value = message.minmax()
        if message["endStep"] == 0:
            report = Report(f"{__class__.__name__}")
            report.add(
                AssertTrue(
                    bool(min_value == 0) and bool(max_value == 0),
                    "min_value == 0 and max_value == 0",
                )
            )

        return super()._from_start(message, p).add(report)

    def _statistical_process(self, message, p) -> Report:
        report = Report("Uerra Statistical Process")

        topd = message.get("typeOfProcessedData", int)
        if topd in [0, 1]:  # Analysis, Forecast
            report.add(IsIn(message["productDefinitionTemplateNumber"], [8, 11]))
        elif topd == 2:  # Analysis and forecast products
            pass
        elif topd in [3, 4]:  # Control forecast products, Perturbed forecast products
            report.add(Eq(message["productDefinitionTemplateNumber"], 61))
        else:
            report.add(Fail(f"Unsupported typeOfProcessedData {topd}"))
            return report

        #  forecastTime for uerra might be all steps decreased by 1 i.e 0,1,2,3,4,5,8,11...29 too many... */
        if message["indicatorOfUnitOfTimeRange"] == 1:
            report.add(Le(message["forecastTime"], 30))

        report.add(Eq(message["timeIncrementBetweenSuccessiveFields"], 0))

        stream = message.get("stream", str)
        if stream != "moda":
            report.add(IsIn(message["endStep"], [1, 2, 4, 5]) | IsMultipleOf(message["endStep"], 3))

        return super()._statistical_process(message, p).add(report)

    def _point_in_time(self, message, p) -> Report:
        report = Report("Uerra Point in Time")

        topd = message.get("typeOfProcessedData", int)
        if topd in [0, 1]:  # Analysis, Forecast
            report.add(IsIn(message["productDefinitionTemplateNumber"], [0, 1]))
        elif topd == 2:  # Analysis and forecast products
            pass
        elif topd == 3:  # Control forecast products
            report.add(Eq(message["productDefinitionTemplateNumber"], 1))
        elif topd == 4:  # Perturbed forecast products
            # Is there always cf in tigge global datasets??
            report.add(
                Le(
                    message["perturbationNumber"],
                    message["numberOfForecastsInEnsemble"] - 1,
                )
            )
        else:
            report.add(Fail(f"Unsupported typeOfProcessedData {topd}"))

        if message["indicatorOfUnitOfTimeRange"] == 1:  # hourly
            report.add(
                IsIn(message["forecastTime"], [1, 2, 4, 5])
                | IsMultipleOf(message["forecastTime"], 3)
            )

        return super()._point_in_time(message, p).add(report)

    def _pressure_level(self, message, p) -> Report:
        report = Report("Uerra Pressure Level")
        levels = [
            1000,
            975,
            950,
            925,
            900,
            875,
            850,
            825,
            800,
            750,
            700,
            600,
            500,
            400,
            300,
            250,
            200,
            150,
            100,
            70,
            50,
            30,
            20,
            10,
        ]
        report.add(IsIn(message["level"], levels, "valid pressure level"))
        return report

    def _height_level(self, message, p) -> Report:
        report = Report("Uerra Height Level")
        levels = [15, 30, 50, 75, 100, 150, 200, 250, 300, 400, 500]
        report.add(IsIn(message["level"], levels, "valid height level"))
        return report



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


class TermColor:
    """
    ANSIescapecodesforterminalcolors.
    """

    # ANSI escape codes for text colors
    BLACK = "\033[30m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"

    # ANSI escape codes for background colors
    BG_BLACK = "\033[40m"
    BG_RED = "\033[41m"
    BG_GREEN = "\033[42m"
    BG_YELLOW = "\033[43m"
    BG_BLUE = "\033[44m"
    BG_MAGENTA = "\033[45m"
    BG_CYAN = "\033[46m"
    BG_WHITE = "\033[47m"

    # Bright colors
    BRIGHT_BLACK = "\033[90m"
    BRIGHT_RED = "\033[91m"
    BRIGHT_GREEN = "\033[92m"
    BRIGHT_YELLOW = "\033[93m"
    BRIGHT_BLUE = "\033[94m"
    BRIGHT_MAGENTA = "\033[95m"
    BRIGHT_CYAN = "\033[96m"
    BRIGHT_WHITE = "\033[97m"

    # Bright background colors
    BRIGHT_BG_BLACK = "\033[100m"
    BRIGHT_BG_RED = "\033[101m"
    BRIGHT_BG_GREEN = "\033[102m"
    BRIGHT_BG_YELLOW = "\033[103m"
    BRIGHT_BG_BLUE = "\033[104m"
    BRIGHT_BG_MAGENTA = "\033[105m"
    BRIGHT_BG_CYAN = "\033[106m"
    BRIGHT_BG_WHITE = "\033[107m"

    # ANSI escape code to reset color│
    RESET = "\033[0m"

    # ANSI escape codes for text styles
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    REVERSED = "\033[7m"

    OKGREEN = BRIGHT_GREEN
    OKCYAN = BRIGHT_MAGENTA
    OKBLUE = BRIGHT_BLUE
    FAIL = BRIGHT_RED
    ENDC = RESET
    SEP = WHITE



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

import numpy as np

from grib_check.Assert import (
    AssertTrue,
    Eq,
    EqDbl,
    Exists,
    Fail,
    Ge,
    IsIn,
    Le,
    Lt,
    Missing,
    Ne,
    Pass,
)
from grib_check.CheckEngine import CheckEngine
from grib_check.Grib import get_gaussian_latitudes
from grib_check.Report import Report


class GeneralChecks(CheckEngine):
    def __init__(self, lookup_table, check_limits=False, check_validity=True):
        super().__init__(lookup_table)
        self.logger = logging.getLogger(__class__.__name__)

        self.register_checks(
            {
                "basic_checks_2": self._basic_checks_2,
                "basic_checks": self._basic_checks,
                "daily_average": self._daily_average,
                "from_start": self._from_start,
                "given_level": self._given_level,
                "given_thickness": self._given_thickness,
                "has_bitmap": self._has_bitmap,
                "has_soil_layer": self._has_soil_layer,
                "has_soil_level": self._has_soil_level,
                "height_level": self._height_level,
                "point_in_time": self._point_in_time,
                "potential_temperature_level": self._potential_temperature_level,
                "potential_vorticity_level": self._potential_vorticity_level,
                "predefined_level": self._predefined_level,
                "predefined_thickness": self._predefined_thickness,
                "resolution_s2s": self._resolution_s2s,
                "resolution_s2s_ocean": self._resolution_s2s_ocean,
                "since_prev_pp": self._since_prev_pp,
                "six_hourly": self._six_hourly,
                "three_hourly": self._three_hourly,
            }
        )
        self.last_n = 0
        self.values = None
        self.check_limits = check_limits
        self.check_validity = check_validity

    # def register_check(self, name, func):
    #     if name in self._check_map:
    #         raise ValueError(f"Check {name} already registered.")
    #     else:
    #         self._check_map[name] = func
    #
    # def override_check(self, name, func):
    #     if name not in self._check_map:
    #         raise ValueError(f"Check {name} not registered.")
    #     else:
    #         self._check_map[name] = func
    #
    # def get_check(self, name):
    #     if name not in self._check_map:
    #         raise ValueError(f"Check {name} not registered.")
    #     else:
    #         return self._check_map[name]

    def _check_date(self, message, p):
        report = Report("Check Date")
        # todo check for how many years back the reforecast is done? Is it coded in the grib???
        # Check if the date is OK
        date = message["date"]
        # report.add(Ge(message["date"], 20060101))

        report.add(Eq((date / 10000).to_int(), message["year"]))
        report.add(Eq(((date % 10000) / 100).to_int(), message["month"]))
        report.add(Eq((date % 100).to_int(), message["day"]))

        return report

    # not registered in the lookup table
    def _statistical_process(self, message, p) -> Report:
        report = Report("Statistical Process")

        topd = message.get("typeOfProcessedData", int)

        if topd.value() in [0, 1]:  # Analysis, Forecast
            report.add(Eq(message["productDefinitionTemplateNumber"], 8, f"topd={topd}"))
        elif topd == 2:  # Analysis and forecast products
            report.add(Eq(message["productDefinitionTemplateNumber"], 8, f"topd={topd}"))
        elif topd in [3, 4]:  # Control forecast products
            pass
        else:
            report.add(Fail(f"Unsupported typeOfProcessedData {topd}"))
            return report

        report.add(Eq(message["numberOfMissingInStatisticalProcess"], 0))
        report.add(Eq(message["typeOfTimeIncrement"], 2))
        # report.add(Eq(message["indicatorOfUnitOfTimeForTheIncrementBetweenTheSuccessiveFieldsUsed"], 255))
        report.add(Eq(message["minuteOfEndOfOverallTimeInterval"], 0))
        report.add(Eq(message["secondOfEndOfOverallTimeInterval"], 0))

        report.add(Eq(message["numberOfTimeRange"], 1))

        if message["indicatorOfUnitForTimeRange"] == 11:
            # Six hourly is OK
            report.add(
                Eq(
                    message["lengthOfTimeRange"] * 6 + message["startStep"],
                    message["endStep"],
                )
            )

        elif message["indicatorOfUnitForTimeRange"] == 10:
            # Three hourly is OK
            report.add(
                Eq(
                    message["lengthOfTimeRange"] * 3 + message.get("startStep", int),
                    message.get("endStep", int)
                )
            )
        else:
            report.add(Eq(message["indicatorOfUnitForTimeRange"], 1))
            report.add(
                Eq(
                    message["lengthOfTimeRange"] + message.get("startStep", int),
                    message.get("endStep", int),
                )
            )

        return report

    # not registered in the lookup table
    def _check_range(self, message, p, min_value=None, max_value=None):
        report = Report("Range check")

        if self.check_limits:
            # See ECC-437
            missing = message.get("missingValue", float)
            if min_value is None or max_value is None:
                min_value, max_value = message.minmax()

            for entry in p["expected"]:
                if entry["key"] == "values":
                    mi1 = entry["min"][0]
                    mi2 = entry["min"][1]
                    ma1 = entry["max"][0]
                    ma2 = entry["max"][1]

                    if (
                        not message["bitMapIndicator"] == 0
                        and min_value != missing
                        and max_value != missing
                    ):
                        if min_value < mi1 or min_value > mi2:
                            report.add(
                                Fail(
                                    f"Minimum value {min_value} is not in range [{mi1}, {mi2}]"
                                )
                            )
                        else:
                            report.add(
                                Pass(
                                    f"Minimum value {min_value} is in range [{mi1}, {mi2}]"
                                )
                            )

                        if max_value < ma1 or max_value > ma2:
                            report.add(
                                Fail(
                                    f"Maximum value {max_value} is not in range [{ma1}, {ma2}]]"
                                )
                            )
                        else:
                            report.add(
                                Pass(
                                    f"Maximum value {max_value} is in range [{ma1}, {ma2}]"
                                )
                            )
        else:
            report.add("Check disabled. Use the option -a or --check_limits to enable it.")
        return report

    # not registered in the lookup table
    def _gaussian_grid(self, message):
        report = Report("Gaussian grid")

        tolerance = 1.0 / 1000000.0  # angular tolerance for grib2: micro degrees
        n = message["numberOfParallelsBetweenAPoleAndTheEquator"]  # This is the key N

        north = message.get("latitudeOfFirstGridPointInDegrees", float)
        south = message.get("latitudeOfLastGridPointInDegrees", float)

        west = message.get("longitudeOfFirstGridPointInDegrees", float)
        east = message.get("longitudeOfLastGridPointInDegrees", float)

        if Ne(n, self.last_n):
            try:
                self.values = get_gaussian_latitudes(n.value())
            except TypeError as e:
                raise e
            except Exception as e:
                report.add(
                    Fail(
                        f"Error: Cannot get gaussian latitudes for N{n.value()}, {str(e)}"
                    )
                )
                self.last_n = 0
                return report
            self.last_n = n

        # TODO
        if self.values is None:
            assert 0
            return report

        if self.values is not None:
            self.values[0] = np.rint(self.values[0] * 1e6) / 1e6

        report.add(
            EqDbl(north, self.values[0], tolerance)
            | EqDbl(south, -self.values[0], tolerance)
        )
        report.add(EqDbl(north, self.values[0], tolerance, "north == self.values[0]"))
        report.add(EqDbl(south, -self.values[0], tolerance, "south == -self.values[0]"))

        if message.is_missing("numberOfPointsAlongAParallel"):  # same as key Ni
            # If missing, this is a REDUCED gaussian grid
            MAXIMUM_RESOLUTION = 640
            report.add(Ne(message["PLPresent"], 0))  # TODO: check this
            report.add(EqDbl(west, 0.0, tolerance, "west == 0.0"))
            report.add(
                Le(
                    n,
                    MAXIMUM_RESOLUTION,
                    f"Gaussian number N (={n}) cannot exceed {MAXIMUM_RESOLUTION}",
                )
            )
        else:
            # REGULAR gaussian grid
            l_west = message["longitudeOfFirstGridPoint"]
            l_east = message["longitudeOfLastGridPoint"]
            parallel = message["numberOfPointsAlongAParallel"]
            we = message["iDirectionIncrement"]
            dwest = message.get("longitudeOfFirstGridPointInDegrees", float)
            deast = message.get("longitudeOfLastGridPointInDegrees", float)
            dwe = message.get("iDirectionIncrementInDegrees", float)
            # printf("parallel=%ld east=%ld west=%ld we=%ld",parallel,east,west,we)

            report.add(
                Eq(
                    parallel,
                    (l_east - l_west) / we + 1,
                    "parallel == (l_east - l_west) / we + 1",
                )
            )
            report.add(
                Lt(
                    ((deast - dwest) / dwe + 1 - parallel).abs(),
                    1e-10,
                    "abs((deast-dwest)/dwe + 1 - parallel) < 1e-10",
                )
            )
            report.add(Eq(message["PLPresent"], 0, "not message.get('PLPresent')"))

        report.add(Ne(message["Nj"], 0))

        if message["PLPresent"] != 0:
            count = message.get_size("pl")
            expected_lon2 = 0
            total = 0
            max_pl = 0
            numberOfValues = message["numberOfValues"]
            numberOfDataPoints = message["numberOfDataPoints"]

            pl = message.get_double_array("pl")

            report.add(
                AssertTrue(len(pl) == count, f"len(pl)({len(pl)}) == count({count})")
            )
            report.add(
                AssertTrue(len(pl) == n * 2, f"len(pl)({len(pl)}) == 2 * n({n})")
            )

            total = 0
            max_pl = pl[0]  # max elem of pl array = num points at equator

            for p in pl:
                total = total + p
                if p > max_pl:
                    max_pl = p

            # Do not assume maximum of pl array is 4N! not true for octahedral
            expected_lon2 = 360.0 - 360.0 / max_pl

            if not EqDbl(east, expected_lon2, tolerance):
                report.add(
                    Fail(
                        f"east actual={east} expected={expected_lon2} diff={expected_lon2-east}"
                    )
                )

            report.add(EqDbl(east, expected_lon2, tolerance, "expected_lon2 == east"))
            report.add(
                Eq(
                    message["numberOfDataPoints"],
                    total,
                    f"GAUSS numberOfValues={numberOfValues} numberOfDataPoints={numberOfDataPoints} sum(pl)={total}",
                )
            )
            report.add(Missing(message, "iDirectionIncrement"))
            report.add(Missing(message, "iDirectionIncrementInDegrees"))
            report.add(Eq(message["iDirectionIncrementGiven"], 0))
            report.add(Eq(message["jDirectionIncrementGiven"], 1))

        report.add(Eq(message["resolutionAndComponentFlags1"], 0))
        report.add(Eq(message["resolutionAndComponentFlags2"], 0))
        report.add(Eq(message["resolutionAndComponentFlags6"], 0))
        report.add(Eq(message["resolutionAndComponentFlags7"], 0))
        report.add(Eq(message["resolutionAndComponentFlags8"], 0))

        return report

    # not registered in the lookup table
    def _latlon_grid(self, message):
        report = Report("latlon grid")

        # tolerance = 1.0/1000000.0 # angular tolerance for grib2: micro degrees
        data_points = message["numberOfDataPoints"]
        meridian = message["numberOfPointsAlongAMeridian"]
        parallel = message["numberOfPointsAlongAParallel"]

        north = message["latitudeOfFirstGridPoint"]
        south = message["latitudeOfLastGridPoint"]
        west = message["longitudeOfFirstGridPoint"]
        east = message["longitudeOfLastGridPoint"]

        # ns= message["jDirectionIncrement"]
        # we= message["iDirectionIncrement"]

        dnorth = message.get("latitudeOfFirstGridPointInDegrees", float)
        dsouth = message.get("latitudeOfLastGridPointInDegrees", float)
        dwest = message.get("longitudeOfFirstGridPointInDegrees", float)
        deast = message.get("longitudeOfLastGridPointInDegrees", float)

        # dns = message.get("jDirectionIncrementInDegrees", float)
        # dwe = message.get("iDirectionIncrementInDegrees", float)

        if message["basicAngleOfTheInitialProductionDomain"] == 0:
            report.add(Missing(message, "subdivisionsOfBasicAngle"))
        else:
            # long basic    = get(h,"basicAngleOfTheInitialProductionDomain")
            # long division = get(h,"subdivisionsOfBasicAngle")
            report.add(Exists(message, "subdivisionsOfBasicAngle"))
            report.add(Ne(message["subdivisionsOfBasicAngle"], 0))

        if message.is_missing("subdivisionsOfBasicAngle"):
            report.add(Eq(message["basicAngleOfTheInitialProductionDomain"], 0))

        report.add(
            Eq(meridian * parallel, data_points, "meridian * parallel == data_points")
        )

        report.add(Eq(message["resolutionAndComponentFlags1"], 0))
        report.add(Eq(message["resolutionAndComponentFlags2"], 0))
        report.add(Eq(message["resolutionAndComponentFlags6"], 0))
        report.add(Eq(message["resolutionAndComponentFlags7"], 0))
        report.add(Eq(message["resolutionAndComponentFlags8"], 0))

        report.add(Eq(message["iDirectionIncrementGiven"], 1))
        report.add(Eq(message["jDirectionIncrementGiven"], 1))
        # https://jira.ecmwf.int/browse/SD-39816
        # 96-97 codeflag resolutionAndComponentFlags = 48 [00110000:(3=1) i direction increments given;(4=1) j direction increments given;(5=0)...
        report.add(Eq(message["resolutionAndComponentFlags"], 48))

        report.add(Eq(message["numberOfOctectsForNumberOfPoints"], 0))
        report.add(Eq(message["interpretationOfNumberOfPoints"], 0))

        if message["iScansNegatively"] != 0:
            tmp = east
            dtmp = deast

            east = west
            west = tmp

            deast = dwest
            dwest = dtmp

        if message["jScansPositively"] != 0:
            tmp = north
            dtmp = dnorth

            north = south
            south = tmp

            dnorth = dsouth
            dsouth = dtmp

        # GRIB2 requires longitudes are always positive */
        report.add(Ge(east, 0, "east >= 0"))
        report.add(Ge(west, 0, "west >= 0"))

        # printf("meridian=%ld north=%ld south=%ld ns=%ld ",meridian,north,south,ns)
        # printf("meridian=%ld north=%f south=%f ns=%f ",meridian,dnorth,dsouth,dns)
        # printf("parallel=%ld east=%ld west=%ld we=%ld ",parallel,east,west,we)
        # printf("parallel=%ld east=%f west=%f we=%f ",parallel,deast,dwest,dwe)

        return report

    # not registered in the lookup table
    def _check_packing(self, message):
        # ECC-1009: Warn if not using simple packing
        report = Report("Check packing")
        report.add(Eq(message["packingType"], "grid_simple"))
        return report

    # not registered in the lookup table
    def _check_validity_datetime(self, message):
        # If we just set the stepRange (for non-instantaneous fields) to its
        # current value, then this causes the validity date and validity time
        # keys to be correctly computed.
        # Then we can compare the previous (possibly wrongly coded) value with
        # the newly computed one

        report = Report("Check Validity Datetime")
        stepType = message.get("stepType", str)

        if stepType != "instant":  # not instantaneous
            # Check only applies to accumulated, max etc.
            # stepRange = message.get("stepRange", str)

            saved_validityDate = message["validityDate"]
            saved_validityTime = message["validityTime"]

            # message.set("stepRange", stepRange)

            validityDate = message["validityDate"]
            validityTime = message["validityTime"]
            if validityDate != saved_validityDate or validityTime != saved_validityTime:
                # print("warning: %s, field %d [%s]: invalid validity Date/Time (Should be %ld and %ld)" % (cfg['filename'], cfg['field'], cfg['param'], validityDate, validityTime))
                report.add(
                    Fail(
                        f"Invalid validity Date/Time (Should be {validityDate} and {validityTime})"
                    )
                )
                # cfg['warning'] += 1

        return report

    def _basic_checks_2(self, message, p):
        report = Report("Basic Checks 2")
        # 2 = analysis or forecast , 3 = control forecast, 4 = perturbed forecast
        report.add(IsIn(message["typeOfProcessedData"], [2, 3, 4]))
        return report

    def _basic_checks(self, message, p):
        report = Report("Basic checks")
        report.add(Eq(message["editionNumber"], 2))
        report.add(Missing(message, "reserved") | Eq(message["reserved"], 0))
        if self.check_validity:
            report.add(Eq(message["isMessageValid"], 1, "Use: grib_get -p isMessageValid file.grib to see the output if you get a failure here."))

        # check min/max ranges
        if self._from_start:
            endStep = message.get("endStep", int)
            if endStep != 0:
                min_value, max_value = message.minmax()
                report.add(self._check_range(message, p, min_value=min_value/endStep, max_value=max_value/endStep))
        else:
            report.add(self._check_range(message, p))

        if self.check_limits:
            values_report = Report("Check values")
            count = 0
            try:
                count = message.get_size("values")
            except Exception as e:
                values_report.add(Fail(f"Cannot get number of values: {e}"))
                return values_report

            values_report.add(Eq(message["numberOfDataPoints"], count))

            try:
                values = message.get_double_array("values")
            except Exception as e:
                values_report.add(Fail(f"Cannot get values: {e}"))
                return values_report

            n = count
            count = len(values)
            if n != count:
                values_report.add(Fail(f"Value count changed {count} -> {n}"))
                return values_report

            report.add(values_report)

        # reports += self._check_packing(message)

        # Section 1

        report.add(Ge(message["gribMasterTablesVersionNumber"], 4))
        report.add(Eq(message["significanceOfReferenceTime"], 1))

        report.add(Eq(message["minute"], 0))
        report.add(Eq(message["second"], 0))
        stream = message.get("stream", str)
        if stream != "moda":
            report.add(Ge(message["startStep"], 0))

        # TODO: validate local usage. Empty for now xxx
        # report.add(Eq(message, "section2.sectionLength", 5)

        # Section 3
        report.add(
            Eq(message["sourceOfGridDefinition"], 0)
        )  # Specified in Code table 3.1

        dtn = message["gridDefinitionTemplateNumber"]

        if dtn in [0, 1]:
            # dtn == 1: rotated latlon
            report.add(self._latlon_grid(message))
        elif dtn == 20:  # Polar stereographic projection
            report.add("Polar stereographic projection checks not implemented!")
        elif dtn == 30:  # Lambert conformal
            report.add("Lambert conformal projection checks not implemented!")
            # lambert_grid(h); # TODO xxx
            # print("warning: Lambert grid - geometry checking not implemented yet!")
            # report.add(Eq(message["scanningMode"], 64)) # M-F data used to have it wrong.. but it might depends on other projection set up as well!
            pass
        elif dtn == 40:  # gaussian grid (regular or reduced)
            report.add(self._gaussian_grid(message))
        else:
            report.add(Fail(f"Unsupported gridDefinitionTemplateNumber {dtn}"))
            return report

        # If there is no bitmap, this should be true
        # CHECK('eq(h,"bitMapIndicator",255)', eq(h,"bitMapIndicator",255))

        if message["bitMapIndicator"] == 255:
            report.add(Eq(message["numberOfValues"], message["numberOfDataPoints"]))
        else:
            report.add(Le(message["numberOfValues"], message["numberOfDataPoints"]))

        # Check values
        report.add(Eq(message["typeOfOriginalFieldValues"], 0))  # Floating point

        report.add(self._check_validity_datetime(message))

        # do not store empty values e.g. fluxes at step 0
        #    todo ?? now it's allowed in the code here!
        #    if not missing(h,"typeOfStatisticalProcessing"):
        #      CHECK('ne(h,"stepRange",0)', ne(h,"stepRange",0))

        return report

    def _daily_average(self, message, p):
        report = Report("Daily Average")
        start_step = message["startStep"]
        end_step = message["endStep"]
        report.add(Eq(start_step, end_step - 24))
        min_value, max_value = message.minmax()
        if end_step == 0:
            report.add(
                AssertTrue(
                    min_value == 0 and max_value() == 0,
                    "min_value == 0 and max_value == 0",
                )
            )
        report.add(self._statistical_process(message, p))

        return report

    def _from_start(self, message, p):
        report = Report("From Start")
        report.add(Eq(message["startStep"], 0))
        report.add(self._statistical_process(message, p))

        return report

    def _point_in_time(self, message, p):
        report = Report("Point in time")
        topd = message.get("typeOfProcessedData", int)
        if topd in [0, 1]:  # Analysis, Forecast
            pass
        elif topd == 2:  # Analysis and forecast products
            report.add(
                Eq(message["productDefinitionTemplateNumber"], 0, f"topd={topd}")
            )
        elif topd == 3:  # Control forecast products
            report.add(Eq(message["perturbationNumber"], 0, f"topd={topd}"))
            report.add(Ne(message["numberOfForecastsInEnsemble"], 0, f"topd={topd}"))
        elif topd == 4:  # Perturbed forecast products
            report.add(Ne(message["perturbationNumber"], 0, f"topd={topd}"))
            report.add(Ne(message["numberOfForecastsInEnsemble"], 0, f"topd={topd}"))
        else:
            report.add(Fail(f"Unsupported typeOfProcessedData {topd}"))

        return report

    def _given_thickness(self, message, p):
        report = Report("Given thickness")
        report.add(Ne(message["typeOfSecondFixedSurface"], 255))
        report.add(Exists(message, "scaleFactorOfSecondFixedSurface"))
        report.add(Exists(message, "scaledValueOfSecondFixedSurface"))

        report.add(Ne(message["typeOfFirstFixedSurface"], 255))
        report.add(Exists(message, "scaleFactorOfFirstFixedSurface"))
        report.add(Exists(message, "scaledValueOfFirstFixedSurface"))
        return report

    def _has_bitmap(self, message, p):
        report = Report("Has bitmap")
        report.add(Eq(message["bitMapIndicator"], 0))
        return report

    def _has_soil_layer(self, message, p):
        report = Report("Has soil layer")
        report.add(Eq(message["topLevel"], message["bottomLevel"] - 1))
        report.add(Le(message["level"], 14))  # max in UERRA
        return report

    def _has_soil_level(self, message, p):
        report = Report("Has soil level")
        report.add(Eq(message["topLevel"], message["bottomLevel"]))
        report.add(Le(message["level"], 14))  # max in UERRA
        return report

    def _height_level(self, message, p):
        report = Report("Height level")
        return report

    def _given_level(self, message, p):
        report = Report("Given level")
        report.add(Ne(message["typeOfFirstFixedSurface"], 255))
        report.add(Exists(message, "scaleFactorOfFirstFixedSurface"))
        report.add(Exists(message, "scaledValueOfFirstFixedSurface"))
        report.add(Eq(message["typeOfSecondFixedSurface"], 255))
        report.add(Missing(message, "scaleFactorOfSecondFixedSurface"))
        report.add(Missing(message, "scaledValueOfSecondFixedSurface"))
        return report

    def _potential_temperature_level(self, message, p):
        report = Report("Potential temperature level")
        report.add(
            Eq(
                message["level"],
                320,
                f'invalid potential temperature level {message["level"]}',
            )
        )
        return report

    def _potential_vorticity_level(self, message, p):
        report = Report("Potential vorticity level")
        report.add(
            Eq(
                message["level"],
                2,
                f'invalid potential vorticity level {message["level"]}',
            )
        )
        return report

    def _predefined_level(self, message, p):
        report = Report("Predefined level")
        report.add(Ne(message["typeOfFirstFixedSurface"], 255))
        report.add(Missing(message, "scaleFactorOfFirstFixedSurface"))
        report.add(Missing(message, "scaledValueOfFirstFixedSurface"))
        report.add(Eq(message["typeOfSecondFixedSurface"], 255))
        report.add(Missing(message, "scaleFactorOfSecondFixedSurface"))
        report.add(Missing(message, "scaledValueOfSecondFixedSurface"))
        return report

    def _predefined_thickness(self, message, p):
        report = Report("Predefined thickness")
        report.add(Ne(message["typeOfFirstFixedSurface"], 255))
        report.add(Missing(message, "scaleFactorOfFirstFixedSurface"))
        report.add(Missing(message, "scaledValueOfFirstFixedSurface"))
        report.add(Ne(message["typeOfSecondFixedSurface"], 255))
        report.add(Missing(message, "scaleFactorOfSecondFixedSurface"))
        report.add(Missing(message, "scaledValueOfSecondFixedSurface"))
        return report

    def _resolution_s2s(self, message, p):
        report = Report("Resolution S2S")
        report.add(Eq(message["iDirectionIncrement"], 1500000))
        report.add(Eq(message["jDirectionIncrement"], 1500000))
        return report

    def _resolution_s2s_ocean(self, message, p):
        report = Report("Resolution S2S Ocean")
        report.add(Eq(message["iDirectionIncrement"], 1000000))
        report.add(Eq(message["jDirectionIncrement"], 1000000))
        return report

    def _since_prev_pp(self, message, p):
        report = Report("Since previous post-processing")
        report.add(Eq(message["indicatorOfUnitForTimeRange"], 1))
        report.add(
            Eq(message["endStep"], message["startStep"] + message["lengthOfTimeRange"])
        )
        report.add(self._statistical_process(message, p))
        return report

    def _six_hourly(self, message, p):
        report = Report("Six hourly")
        if message["indicatorOfUnitForTimeRange"] == 11:
            report.add(Eq(message["lengthOfTimeRange"], 1))
        else:
            report.add(Eq(message["lengthOfTimeRange"], 6))
        report.add(Eq(message["endStep"], message["startStep"] + 6))

        report.add(self._statistical_process(message, p))
        return report

    def _three_hourly(self, message, p):
        report = Report("Three hourly")
        if message["indicatorOfUnitForTimeRange"] == 11:
            report.add(Eq(message["lengthOfTimeRange"], 1))
        else:
            report.add(Eq(message["lengthOfTimeRange"], 3))
        report.add(Eq(message["endStep"], message["startStep"] + 3))

        report.add(self._statistical_process(message, p))
        return report



#!/usr/bin/env python3

#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import argparse
import logging
import multiprocessing
import os
import signal
import sys

from .checker.Crra import Crra
from .checker.Lam import Lam
from .checker.S2S import S2S
from .checker.S2SRefcst import S2SRefcst
from .checker.Tigge import Tigge
from .checker.Uerra import Uerra
from .checker.Wpmip import Wpmip
from .FileScanner import FileScanner
from .Grib import Grib
from .LookupTable import SimpleLookupTable
from .Message import Message
from .Report import Report
from .ValueFormat import formatter

signal.signal(signal.SIGINT, lambda x, y: sys.exit(0))  # Disable traceback on Ctrl+C


def worker(filename, message_buffer, pos, checker, args):
    message = Message(message_buffer=message_buffer, position=pos)

    sub_report = Report(f"field {message.position()}")
    sub_report.add(checker.validate(message))

    report = Report(f"{filename}")
    report.add(sub_report)

    print(
        report.as_string(
            max_level=args.report_depth,
            color=args.color,
            failed_only=args.failed_only,
            output_type=args.output_type,
        ),
        end="",
        flush=True,
    )

    # return report
    return None


class GribCheck:
    def __init__(self, args):
        self.args = args
        self.logger = logging.getLogger(__class__.__name__)

    def run(self):
        """
        lam: local area model
        s2s: subseasonal to subseasonal
        s2s_refcst: subseasonal to subseasonal reforecast
        uerra: uncertainty estimation reanalysis
        crra: climate reanalysis
        """
        script_path = os.path.dirname(os.path.realpath(__file__))
        tigge_params = (
            self.args.parameters
            if self.args.parameters is not None
            else f"{script_path}/checker/TiggeParameters.jsonnet"
        )
        wpmip_params = (
            self.args.parameters
            if self.args.parameters is not None
            else f"{script_path}/checker/WpmipParameters.jsonnet"
        )
        crra_params = (
            self.args.parameters
            if self.args.parameters is not None
            else f"{script_path}/checker/CrraParameters.jsonnet"
        )

        if self.args.convention == "tigge":
            checker = Tigge(SimpleLookupTable(tigge_params), check_limits=self.args.check_limits, check_validity=self.args.validity_check)
        elif self.args.convention == "wpmip":
            checker = Wpmip(SimpleLookupTable(wpmip_params), check_limits=self.args.check_limits, check_validity=self.args.validity_check)
        elif self.args.convention == "s2s":
            checker = S2S(SimpleLookupTable(tigge_params), check_limits=self.args.check_limits, check_validity=self.args.validity_check)
        elif self.args.convention == "s2s_refcst":
            checker = S2SRefcst(SimpleLookupTable(tigge_params), check_limits=self.args.check_limits, check_validity=self.args.validity_check)
        elif self.args.convention == "uerra":
            checker = Uerra(SimpleLookupTable(tigge_params, ignore_keys=["model"]),
                            check_limits=self.args.check_limits, check_validity=self.args.validity_check,)
        elif self.args.convention == "crra":
            checker = Crra(
                SimpleLookupTable(crra_params, ignore_keys=["model"]),
                check_limits=self.args.check_limits, check_validity=self.args.validity_check,
            )
        elif self.args.convention == "lam":
            checker = Lam(SimpleLookupTable(tigge_params), check_limits=self.args.check_limits, check_validity=self.args.validity_check)
        else:
            raise ValueError("Unknown data type")

        if self.args.num_jobs > 1:
            results = []
            with multiprocessing.Pool(processes=self.args.num_jobs) as pool:
                for filename in FileScanner(self.args.path):
                    grib = Grib(filename)
                    for pos, message in enumerate(grib):
                        results.append(
                            pool.apply_async(
                                worker,
                                (
                                    filename,
                                    message.get_buffer(),
                                    pos + 1,
                                    checker,
                                    self.args,
                                ),
                            )
                        )
                for result in results:
                    result.wait()
        else:
            for filename in FileScanner(self.args.path):
                grib = Grib(filename)
                for pos, message in enumerate(grib):
                    worker(filename, message.get_buffer(), pos + 1, checker, self.args)


def main():
    parser = argparse.ArgumentParser(description="""GribCheck is a tool that validates project-specific conventions of GRIB files.
It performs a set of checks on GRIB messages to ensure they comply with the project's internal standards and expectations.
    """)

    parser.add_argument("path", nargs="+", help="path(s) to a GRIB file(s) or directory(s)", type=str)
    parser.add_argument("-L", "--check-limits", help="check value ranges (min/max limits)", action="store_true")
    parser.add_argument(
        "-C",
        "--convention",
        help="data convention. The following conventions are experimental: wpmip.",
        choices=[
            "tigge",
            "s2s",
            "s2s_refcst",
            "uerra",
            "crra",
            "lam",
            "wpmip",
        ],
        required=True,
        type=str,
    )
    parser.add_argument("-l", "--report-depth", help="report depth", type=int, default=10)
    parser.add_argument("-d", "--debug", help="debug mode", action="store_true")
    parser.add_argument("-p", "--parameters", help="path to parameters file", default=None)
    parser.add_argument("-c", "--color", help="use color in output", action="store_true")
    parser.add_argument("-j", "--num-jobs", help="number of jobs", type=int, default=1)
    parser.add_argument("-f", "--failed-only", help="show only failed checks", action="store_true")
    parser.add_argument("-o", "--output-type", help="output format", choices=["short", "tree"], default="tree")
    parser.add_argument("-v", "--version", action="version", version="%(prog)s 0.0.3")
    parser.add_argument("-t", "--show-type", help="show value type", action="store_true")
    parser.add_argument(
        "--validity-check",
        help='check validity of messages using the "isMessageValid" key provided by ecCodes. (experimental)',
        action="store_true",
    )
    args = parser.parse_args()

    if args.debug:
        print("Debug mode")
        logging.basicConfig(
            filename="grib_check.log",
            format="%(asctime)s %(name)s %(levelname)-8s %(process)d %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
            level=logging.DEBUG,
        )

    logger = logging.getLogger(__name__)
    logger.info("Started")
    if args.show_type:
        formatter.set_format("{}:{}", show_type=True)

    grib_check = GribCheck(args)
    return grib_check.run()


if __name__ == "__main__":
    ret = main()
    # sys.exit(0 if ret is None else ret)



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from grib_check.Assert import Eq, Fail, IsIn, IsMultipleOf, Le, Ne, Pass
from grib_check.KeyValue import KeyValue
from grib_check.Report import Report

from .GeneralChecks import GeneralChecks


class Tigge(GeneralChecks):
    def __init__(self, lookup_table, check_limits=False, check_validity=True):
        super().__init__(lookup_table, check_limits=check_limits, check_validity=check_validity)
        self.logger = logging.getLogger(__class__.__name__)
        self.register_checks(
            {
                "pressure_level": self._pressure_level,
            }
        )

    def _basic_checks(self, message, p):
        report = Report("Tigge Basic Checks")
        report.add(Eq(message["versionNumberOfGribLocalTables"], 0))
        # Only 00, 06 12 and 18 Cycle OK
        report.add(IsIn(message["hour"], [0, 6, 12, 18]))
        report.add(IsIn(message["productionStatusOfProcessedData"], [4, 5]))
        report.add(Le(message["endStep"], 30 * 24))
        report.add(IsMultipleOf(message["step"], 6))
        report.add(self._check_date(message, p))

        return super()._basic_checks(message, p).add(report)

    def _pressure_level(self, message, p):
        report = Report("TIGGE Pressure level")
        levels = [1000, 200, 250, 300, 500, 700, 850, 925, 50]
        report.add(IsIn(message["level"], levels))
        return report

    # not registered in the lookup table
    def _statistical_process(self, message, p) -> Report:
        report = Report("Tigge Statistical Process")

        topd = message.get("typeOfProcessedData", int)

        if topd in [0, 1, 2]:  # Analysis, Forecast, Analysis and forecast products
            pass
        elif topd in [3, 4]:  # Control forecast products, Perturbed forecast products
            report.add(Eq(message["productDefinitionTemplateNumber"], 11))
        else:
            report.add(Fail(f"Unsupported typeOfProcessedData {topd}"))
            return report

        if message.get("indicatorOfUnitOfTimeRange") == 11:  # six hours
            # Six hourly is OK
            pass
        else:
            report.add(Eq(message["indicatorOfUnitOfTimeRange"], 1))
            report.add(IsMultipleOf(message["forecastTime"], 6))

        report.add(Eq(message["timeIncrementBetweenSuccessiveFields"], 0))
        report.add(IsMultipleOf(message["endStep"], 6))

        return super()._statistical_process(message, p).add(report)

    def _from_start(self, message, p) -> Report:
        report = Report("Tigge From Start")
        endStep = message["endStep"]
        if endStep == 0:
            min_value, max_value = message.minmax()
            if min_value == 0 and max_value == 0:
                report.add(Pass(f"min and max are both {KeyValue(None, 0)} for {endStep}"))
            else:
                report.add(Fail(f"min and max should both be {KeyValue(None, 0)} for {endStep} but are {KeyValue(None, min_value)} and {KeyValue(None, max_value)}"))

        return super()._from_start(message, p).add(report)

    def _point_in_time(self, message, p) -> Report:
        report = Report("Tigge Point in Time")

        topd = message.get("typeOfProcessedData", int)
        if topd in [0, 1]:  # Analysis, Forecast
            if message.get("productDefinitionTemplateNumber") == 1:
                report.add(
                    Ne(message["numberOfForecastsInEnsemble"], 0, f"topd = {topd}")
                )
                report.add(
                    Le(
                        message["perturbationNumber"],
                        message.get("numberOfForecastsInEnsemble"),
                        f"topd = {topd}",
                    )
                )
        elif topd == 2:  # Analysis and forecast products
            pass
        elif topd == 3:  # Control forecast products
            report.add(
                Eq(message["productDefinitionTemplateNumber"], 1, f"topd = {topd}")
            )
        elif topd == 4:  # Perturbed forecast products
            report.add(
                Eq(message["productDefinitionTemplateNumber"], 1, f"topd = {topd}")
            )
            report.add(
                Le(
                    message["perturbationNumber"],
                    message["numberOfForecastsInEnsemble"] - 1,
                    f"topd = {topd}",
                )
            )
        else:
            report.add(Fail(f"Unsupported typeOfProcessedData {topd}"))

        return super()._point_in_time(message, p).add(report)

    def _height_level(self, message, p) -> Report:
        report = Report("Tigge Height Level")
        report.add(Fail("Not implemented: dummy height_level()"))
        return report



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

import numpy as np
from eccodes import (
    codes_get,
    codes_get_double_array,
    codes_get_long_array,
    codes_get_message,
    codes_get_size,
    codes_is_missing,
    codes_keys_iterator_get_name,
    codes_keys_iterator_new,
    codes_keys_iterator_next,
    codes_new_from_message,
    codes_release,
    codes_set,
)

from .Assert import Eq
from .KeyValue import KeyValue
from .Report import Report


class Message:
    def __init__(self, handle=None, message_buffer=None, position=None):
        self.__h = None
        # assert position != 0
        if position is not None and position < 1:
            raise Exception("Position not supported")
        assert handle is not None or message_buffer is not None
        assert (handle is not None and message_buffer is not None) is not True
        self.logger = logging.getLogger(__class__.__name__)

        if handle is not None:
            self.__h = handle
        elif message_buffer is not None:
            self.__h = codes_new_from_message(message_buffer)

        self.__position = position
        self.min = None
        self.max = None

    def __getitem__(self, key):
        return self.get(key)

    def get_report(self):
        report = Report("Message dump")
        iterator = codes_keys_iterator_new(self.__h, "ls")
        while iterator is not None and codes_keys_iterator_next(iterator):
            key = codes_keys_iterator_get_name(iterator)
            report.add(f"{key} = {self[key].value()}")
        report.add("")

        iterator = codes_keys_iterator_new(self.__h, "mars")
        while iterator is not None and codes_keys_iterator_next(iterator):
            key = codes_keys_iterator_get_name(iterator)
            report.add(f"{key} = {self[key].value()}")
        report.add("")

        report.add(f"model = {self['model'].value()}")
        report.add(f"paramId = {self['paramId'].value()}")
        report.add(f"discipline = {self['discipline'].value()}")
        report.add(f"parameterCategory = {self['parameterCategory'].value()}")
        report.add(f"parameterNumber = {self['parameterNumber'].value()}")
        report.add(
            f"typeOfStatisticalProcessing = {self['typeOfStatisticalProcessing'].value()}"
        )
        report.add(
            f"typeOfFirstFixedSurface = {self['typeOfFirstFixedSurface', int].value()}"
        )

        return report

    @property
    def handle(self):
        return self.__h

    def __del__(self):
        if self.__h is not None:
            codes_release(self.__h)

    def set(self, key, value):
        try:
            codes_set(self.__h, key, value)
        except Exception:
            self.logger.debug(f"KeyError: {key}={value}")

    def get(self, key, datatype=None) -> KeyValue:
        try:
            return KeyValue(key, codes_get(self.__h, key, ktype=datatype))
        except Exception:
            self.logger.debug(f"KeyError: {key}")
            return KeyValue(key, None)

    def get_size(self, key) -> int:
        return codes_get_size(self.__h, key)

    def get_double_array(self, key) -> np.ndarray:
        return codes_get_double_array(self.__h, key)

    def get_long_array(self, key) -> np.ndarray:
        return codes_get_long_array(self.__h, key)

    def get_buffer(self) -> bytes:
        return codes_get_message(self.__h)

    # Check whether the given key has the value 'missing'
    def is_missing(self, key) -> bool:
        return False if codes_is_missing(self.__h, key) == 0 else True

    def position(self):
        return self.__position

    def minmax(self):
        if self.min is not None and self.max is not None:
            return self.min, self.max

        try:
            values = self.get_double_array("values")
        except Exception:
            return

        bitmap = not Eq(self, "bitMapIndicator", 255)

        if bitmap:
            missing = self.get("missingValue")
            it = (v for v in values if v != missing)
            try:
                m = M = next(it)
            except StopIteration:
                m = M = missing
            else:
                for v in it:
                    if v < m:
                        m = v
                    elif v > M:
                        M = v
        else:
            m = min(values)
            M = max(values)

        self.min, self.max = m, M
        return m, M



#
# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

class ValueFormat:
    def __init__(self, fmt: str = "{}", show_type: bool = False):
        self.__type_map = {
            int: "i",
            float: "d",
            str: "s",
        }
        self.__fmt = fmt
        self.__show_type = show_type

    def set_format(self, fmt: str, show_type: bool = True):
        self.__show_type = show_type
        self.__fmt = fmt

    def format(self, value=5):
        if type(value) is list:
            types = set(self.__type_map[type(v)] for v in value)
            if len(types) == 1:
                type_str = types.pop()
            else:
                type_str = "mixed"
        else:
            type_str = self.__type_map.get(type(value), "?")

        if self.__show_type:
            return self.__fmt.format(value, type_str)
        else:
            return self.__fmt.format(value)


# formatter = ValueFormat("{}:{}", show_type=True)
formatter = ValueFormat("{}", show_type=False)
